import typing

ItemType = typing.TypeVar("ItemType")
KeyType = typing.TypeVar("KeyType")
ValueType = typing.TypeVar("ValueType")
MessageType = typing.TypeVar("MessageType")
ProjectionType = typing.TypeVar("ProjectionType")
AggregatorResultType = typing.TypeVar("AggregatorResultType")
ElementType = typing.TypeVar("ElementType")
BlockingProxyType = typing.TypeVar("BlockingProxyType")
